#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import ipaddress
import json
import logging
import re

from typing import Set, Union, Dict, List
from urllib.parse import urljoin, urlparse

import requests


class PySecurityTXTException(Exception):
    pass


class SecurityTXTNotAvailable(PySecurityTXTException):
    pass


class PySecurityTXT():

    def __init__(self, loglevel: int=logging.INFO):
        """Do things to a security.txt file."""
        self.logger = logging.getLogger(f'{self.__class__.__name__}')
        self.logger.setLevel(loglevel)
        self.session = requests.session()
        self.expected_paths = ['/.well-known/security.txt', '/security.txt']

    def _try_get_url(self, url: str) -> str:
        try:
            response = self.session.get(url)
            response.raise_for_status()
        except (requests.HTTPError, requests.exceptions.ConnectionError):
            raise SecurityTXTNotAvailable(f'Unable to get the file from: {url}')
        return response.text

    def _all_possible_domains(self, hostname: str) -> Set[str]:
        hostname_parts = hostname.split('.')
        if len(hostname_parts) <= 2:
            return {hostname, }
        current_domain = '.'.join(hostname_parts[-2:])
        to_return: Set[str] = {current_domain, }
        for domain_part in reversed(hostname_parts[:-2]):
            current_domain = f'{domain_part}.{current_domain}'
            to_return.add(current_domain)
        return to_return

    def parse(self, file: str) -> Dict[str, Union[str, List[str]]]:
        """Takes a security.txt file, parses it.

            :param file: The security.txt file.
        """
        to_return = {}
        if acknowledgments := re.findall("^[A,a]cknowledgments[:]? (.*)$", file, re.MULTILINE):
            if len(acknowledgments) == 1:
                to_return['acknowledgments'] = acknowledgments[0]
            else:
                to_return['acknowledgments'] = acknowledgments
        if canonical := re.findall("^[C,c]canonical[:]? (.*)$", file, re.MULTILINE):
            if len(canonical) == 1:
                to_return['canonical'] = canonical[0]
            else:
                to_return['canonical'] = canonical
        if contact := re.findall("^[C,c]ontact[:]? (.*)$", file, re.MULTILINE):
            if len(contact) == 1:
                to_return['contact'] = contact[0]
            else:
                to_return['contact'] = contact
        if encryption := re.findall("^[E,e]ncryption[:]? (.*)$", file, re.MULTILINE):
            if len(encryption) == 1:
                to_return['encryption'] = encryption[0]
            else:
                to_return['encryption'] = encryption
        if expires := re.findall("^[E,e]xpires[:]? (.*)$", file, re.MULTILINE):
            if len(expires) == 1:
                to_return['expires'] = expires[0]
            else:
                to_return['expires'] = expires
        if hiring := re.findall("^[H,h]iring[:]? (.*)$", file, re.MULTILINE):
            if len(hiring) == 1:
                to_return['hiring'] = hiring[0]
            else:
                to_return['hiring'] = hiring

        if policy := re.findall("^[P,p]olicy[:]? (.*)$", file, re.MULTILINE):
            if len(policy) == 1:
                to_return['policy'] = policy[0]
            else:
                to_return['policy'] = policy
        if prefered_languages := re.findall("^[P,p]referred-[L,l]anguages[:]? (.*)$", file, re.MULTILINE):
            # this one must be there only once, and contains a list of languages
            # I have limited trust on that, so let's normalize it
            languages: List[str] = []
            for lang_list in prefered_languages:
                languages += [language.strip() for language in lang_list.split(',') if language.strip()]
            to_return['prefered-languages'] = languages
        return to_return

    def get(self, hint: str, /, *, parse: bool=False) -> str:
        '''Get the security.txt file.

            :param hint: It can be a domain, an IP or an URL (and we try to figure out where the file is), or a full URL to the file.
            :param parse: Parse the file and returns a json dictionary with the relevant fields
        '''
        if hint.endswith('security.txt'):
            response = self._try_get_url(hint)
            if parse:
                return json.dumps(self.parse(response))
            return response

        if re.search("^http[s]?://", hint):
            # we have a URL, get the hostname
            splitted_url = urlparse(hint)
            if not (hostname := splitted_url.hostname):
                raise PySecurityTXTException(f'Unable to get a hostname out of the hint: {hint}')
        else:
            hostname = hint

        try:
            ipaddress.ip_address(hostname)
            all_domains = {hostname, }
            ip = True
        except ValueError:
            all_domains = self._all_possible_domains(hostname)
            ip = False
        test_urls: Set[str] = set()
        for domain in all_domains:
            # we have what should be a domain or an ip, let's try a few things
            for expected_path in self.expected_paths:
                test_urls.update([
                    urljoin(f'https://{domain}', expected_path),
                    urljoin(f'http://{domain}', expected_path)
                ])

            if not domain.startswith('www') and not ip:
                for expected_path in self.expected_paths:
                    test_urls.update([
                        urljoin(f'https://www.{domain}', expected_path),
                        urljoin(f'http://www.{domain}', expected_path)
                    ])

        for url in sorted(test_urls, key=len, reverse=True):  # Longest URL first, so /.well-known/ path is prioritized
            try:
                response = self._try_get_url(url)
                if re.search("^[C,c]ontact", response, re.MULTILINE):
                    # Because we often get responses that are just plain HTML pages and a 2XX status code.
                    break
            except SecurityTXTNotAvailable:
                self.logger.debug(f'Not available on {url}')
        else:
            raise SecurityTXTNotAvailable(f'Unable to file on {", ".join(test_urls)}')

        if not parse:
            return response
        return json.dumps(self.parse(response))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pysecuritytxt']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.28.2,<3.0.0']

extras_require = \
{'docs': ['Sphinx>=6.1.3,<7.0.0']}

entry_points = \
{'console_scripts': ['pysecuritytxt = pysecuritytxt:main']}

setup_kwargs = {
    'name': 'pysecuritytxt',
    'version': '1.0.1',
    'description': 'Python CLI and module for querying security.txt files on domains.',
    'long_description': '# Python client and module for querying .well-known/security.txt files\n\nGive it a domain, it tries to fetch the security.txt file\n\n## Installation\n\n```bash\npip install pysecuritytxt\n```\n\n## Usage\n\n### Command line\n\nYou can use the `pysecuritytxt` command:\n\n```bash\nusage: pysecuritytxt [-h] [-p] url_or_domain\n\nTry to get a security.txt file\n\npositional arguments:\n  url_or_domain  Try to get the file from there.\n\noptions:\n  -h, --help     show this help message and exit\n  -p, --parse    Parse the response, returns dict\n```\n\n### Library\n\nSee [API Reference](https://pysecuritytxt.readthedocs.io/en/latest/api_reference.html)\n',
    'author': 'Raphaël Vinot',
    'author_email': 'raphael.vinot@circl.lu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Lookyloo/pysecuritytxt',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
